%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                   DEMO FOR "JOINT MULTIMODAL SPARSE RECOVERY"
%                     Release ver. 1.0  (3 Dec. 2020)
%
%
% The code performs the joint sparse recovery of misaligned multimodal images
% described in the paper:
%
% N. Eslahi and A. Foi, "Joint sparse recovery of misaligned multimodal 
% images via adaptive local and nonlocal cross-modal regularization"
% in Proc. 2019 IEEE 8th International Workshop on Computational Advances 
% in Multi-Sensor Adaptive Processing (CAMSAP), pp. 111-115, Dec. 2019.
%
%                      Authors: Nasser Eslahi & Alessandro Foi
%                           Tampere University, Finland
%                       contact: firstname.lastname@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Path to the functions and data

clear; 
addpath(genpath('./'));

%% Loading signals of interest (+ deforming them)

rng(0,'v4');

experiment        = 1;                                      % 1: multichannel MR recovery, 2: Multichannel sparse recovery

if  experiment == 1  % multichannel sparse recovery
    load RGB_Toy.mat                                        % Taken from CAVE database:
                                                            % http://www.cs.columbia.edu/CAVE/databases/multispectral/
    x1 = x(:,:,1);                                          % taking Red channel as x1
    x2 = x(:,:,3);                                          % taking Blue channel as x2

    defpar_x1     = [0 0 1 0];                              % deformation parameters for x1, i.e. [hor_shift vert_shift zoom rotation]
    defpar_x2     = [5.3 2.8 1.25 10];                      % deformation parameters for x2 
    
else                 % multicontrast sparse MR recovery
    load BrainWeb.mat                                       % T1 and T2 BrainWeb transverse 2D slices 
                                                            % Taken from simulated brain database:
                                                            % https://brainweb.bic.mni.mcgill.ca/

    defpar_x1     = [0 0 1 0];                              % deformation parameters for x1, i.e. [hor_shift vert_shift zoom rotation]
    defpar_x2     = [4.4 5.5 1.2 -6];                       % deformation parameters for x2 
end

affTransf_x1      = Affine_Transf(defpar_x1, size(x1));     % computing the affine transformation matrix w.r.t. defpar_x1
[x1, ~, ~]        = Affine_Deform(x1, affTransf_x1);        % deforming x1

affTransf_x2      = Affine_Transf(defpar_x2, size(x2));     % computing the affine transformation matrix w.r.t. defpar_x2
[x2, ~, ~]        = Affine_Deform(x2, affTransf_x2);        % deforming x2
    
[height1, width1] = size(x1);
[height2, width2] = size(x2);

%% Measurement acquisition using 2 different modalities -- Eq.(1)

if experiment == 1  % multichannel sparse recovery
    [A1, At1, S1] =  Sampling( x1, 35, 'radial' );          % pseudo-radial sampling with 35 radial lines
    [A2, At2, S2] =  Sampling( x2,.15, 'Pseudo_random' );   % variable density pseudo-random sampling

else                % multicontrast sparse MR recovery
    [A1, At1, S1] =  Sampling( x1, 20, 'radial' );          % pseudo-radial sampling with 20 radial lines
    [A2, At2, S2] =  Sampling( x2, 30, 'radial' );          % pseudo-radial sampling with 30 radial lines
end   
snr1 = 30; snr2 = 20;                                       % SNR (dB) -- noise strength

y1                 = A1(x1); 
y1(S1~=0)          = Additive_WGN(y1(S1~=0),snr1);          % adding WGN to the measurement #1

y2                 = A2(x2);
y2(S2~=0)          = Additive_WGN(y2(S2~=0),snr2);          % adding WGN to the measurement #2

x1_b = real(At1(y1));                                       % back-projection (inverse FFT)
x2_b = real(At2(y2));                                       % back-projection (inverse FFT)

%% Computing the PSD of measurement error counterparts w.r.t. the sparsifying transforms used internally by the filter

HT_Tr             = 'bior1.5';                              % the sparsifying transform in the hard thresholding (HT)
                                                            % stage of BM3D filter
blkSize_HT_Tr     = 8;                                      % size of 2D block transform used by the filter (BM3D)
rootPSD_Ad_e      = zeros(blkSize_HT_Tr,blkSize_HT_Tr,2);

for kk = 1:2                                                % number of modalities
    if kk == 1
       y = y1; At = At1; A = A1; X = x1; S = S1;
    else
       y = y2; At = At2; A = A2; X = x2; S = S2;
    end
   
   % Computing the root-PSD of "Ad_e = A^{\dagger} epsilon" using a synthetic error
   eps           = y - A(X);                                % oracle measurement error
   eps           = eps(S~=0);
   
   std_real_eps  = std(real(eps(:)));                       % std of real counterpart
   std_imag_eps  = std(imag(eps(:)));                       % std of imaginary counterpart
   
   hat_eps_r     = randn(size(X));
   hat_eps_r     = (hat_eps_r-mean(hat_eps_r(:)))./std(hat_eps_r(:));
   hat_eps_r     = std_real_eps * hat_eps_r;
   
   hat_eps_i     = randn(size(X));
   hat_eps_i     = (hat_eps_i-mean(hat_eps_i(:)))./std(hat_eps_i(:));
   hat_eps_i     = std_imag_eps * hat_eps_i;
   
   hat_eps       = S.*(hat_eps_r + 1i*hat_eps_i);            % dummy measurement error with the same
                                                             % mean and variance as the oracle measurement error
   
   Ad_e          = At(hat_eps);                             
                                                       
    % computing the root-PSD of measurement error counterpart through sample variance over 
    % 2D HT_Tr spectra of all (blkSize_HT_Tr x blkSize_HT_Tr) blocks extracted from "real(Ad_e)"
    % -- we only filter the real part
    rootPSD_Ad_e(:,:,kk)  = Est_rootPSD(real(Ad_e),...
      [blkSize_HT_Tr blkSize_HT_Tr],{HT_Tr,HT_Tr}, 0, 'sVar'); 

  if kk==1
      Ad_e1       = Ad_e; clear Ad_e
  else
      Ad_e2       = Ad_e;
  end
  
end

clear real_eps imag_eps std_real_eps std_imag_eps hat_eps_r hat_eps_i hat_eps Ad_e A At y X

    % the elements of difference image "x_k - x" in the null space of measurement matrix "A"
    % are ignored in surrogate noise signal (i.e., the signal by which we estimate the noise root-PSD),
    % thus we compute the 2D compensation arrays of the root-PSDs in order to compensate such effect
[Tr_12, ~]        = GetTransformMatrix(blkSize_HT_Tr, HT_Tr);
W1                = zeros(blkSize_HT_Tr);               
W2                = W1;
for j1 = 1:blkSize_HT_Tr
    for j2        = 1:blkSize_HT_Tr
        Z         = zeros(blkSize_HT_Tr);
        Z(j1,j2)  = 1;
        B         = Tr_12'*Z*Tr_12;
        W1(j1,j2) = sum(sum(abs((fft2(B,height1, width1)).*S1).^2))/(height1*width1);
        W2(j1,j2) = sum(sum(abs((fft2(B,height2, width2)).*S2).^2))/(height2*width2);
    end
end
comp1             = 1./sqrt(W1);                            % HT_Tr root-PSD compensation array of A1^{\dagger}A1.
comp2             = 1./sqrt(W2);                            % HT_Tr root-PSD compensation array of A2^{\dagger}A2.

clear B Z Tr_12

%% Separate reconstruction using PnP and BM3D as the denoiser -- Tackling Eq.(2)

PnP_solver_sep    = 'FBS';                                  % 'FBS', 'FISTA', 'ISTA', PnP approach for separate recovery phase
iters             = [100 100];                              % iteration numbers
PSNR_init         = zeros(max(iters),2);
lambda_bm3d       = 2.7;                                    % BM3D thresholding factor

if experiment ==1  % multichannel sparse recovery
    % step size (can be different for each modality)
    if strcmpi(PnP_solver_sep,'FISTA')
        rho       = [.44 1.04];
    elseif strcmpi(PnP_solver_sep,'FBS')
        rho       = [1.89 1.89];
    elseif strcmpi(PnP_solver_sep,'ISTA')
        rho       = [2 2];
    else
        error('Your solver does not exist!')
    end
    
else              % multicontrast sparse MR recovery
    % step size (can be different for each modality and PnP approach)
    if strcmpi(PnP_solver_sep,'FISTA')
        rho       = [1.3 1.3];
    elseif strcmpi(PnP_solver_sep,'FBS')
        rho       = [1.8 1.8];
    elseif strcmpi(PnP_solver_sep,'ISTA')
        rho       = [1.9 1.9];
    else
        error('Your solver does not exist!')
    end
end

for kk = 1:2     %%% Iterative Separate Recovery %%%
    if kk == 1
       y = y1; At = At1; A = A1; comp = comp1; X = x1;
       rootPSD_Ad_ee = rootPSD_Ad_e(:,:,1);
    else
       y = y2; At = At2; A = A2; comp = comp2; X = x2;
       rootPSD_Ad_ee = rootPSD_Ad_e(:,:,2);
    end
    
    alpha_k       = 1;                                     
    r_k           = zeros(size(y));
    u_k           = real(At(y));
    x_k           = u_k;
    fprintf('Separate Recovery of x_%d via PnP-%s\n',kk,PnP_solver_sep);
    for k = 1:iters(kk)

        z_k       = x_k + rho(kk).*real(At(r_k));           % the noisy signal to be filtered, Eq.(4)
        
        if k>1      % Estimating stationary correlated noise root-PSD
            w_k         = real(At(r_k));
            rootPSD_w_k = Est_rootPSD(w_k, [8 8], {HT_Tr,HT_Tr}, 0, 'sVar');
                    % computing the root-PSD of the surrogate noise      
            rootPSD_HT  = sqrt( (rho(kk).*rootPSD_w_k).^2 +...
                abs(rootPSD_w_k.^2-rootPSD_Ad_ee.^2).*comp.^2 );
            rootPSD_Wie = zeros(8);                         % we consider only HT-stage
        else        % the first iteration is treated as AWGN denoising
            rootPSD_HT  = .2*max(z_k(:)).*ones(8);
            rootPSD_Wie = zeros(8);                         % we consider only HT-stage
        end
   
        u_k_1     = u_k;
        % normalizing the noisy image as required by BM3D
        z_k_n     = (z_k - min(z_k(:)))./(max(z_k(:))-min(z_k(:)));
        % rescaling of the root-PSD (since z_k is rescaled)
        rPSD_HT   = rootPSD_HT/(max(z_k(:))-min(z_k(:)));       
        rPSD_Wie  = rootPSD_Wie/(max(z_k(:))-min(z_k(:)));
        
        u_k       = BM3D_colored(z_k_n, rPSD_HT, rPSD_Wie, lambda_bm3d, 0,'np'); % BM3D filtering, Eq.(3b)
        
        u_k       = (u_k.*(max(z_k(:))-min(z_k(:))))+min(z_k(:));

        alpha_k_1 = alpha_k;
        alpha_k   = .5 + .5*sqrt(1+4*alpha_k^2);
        
        if strcmpi(PnP_solver_sep,'FISTA')
            t_k   = (alpha_k_1-1)/alpha_k;
        elseif strcmpi(PnP_solver_sep,'FBS')
            if k<160, t_k   = (alpha_k^k-1)/(alpha_k^(k+1)); else, t_k = 0; end
        elseif strcmpi(PnP_solver_sep,'ISTA')
            t_k   = 0;
        end
        x_k       = u_k + t_k.*(u_k-u_k_1);                 % updating x_k, Eq.(3c)
        r_k       = y - A(x_k);                             % updating measurement residual r_k, used in Eq.(3a)

        PSNR_init(k,kk) = 10*log10((max(X(:))).^2/mean((X(:)-x_k(:)).^2));
        fprintf('Iteration = %d, PSNR = %0.3f dB\n', k, PSNR_init(k,kk));   

    end
    
    if kk==1
        x1_0 = x_k;                                         % the final estimate of x1 by separate recovery
        r1_k = r_k;                                         % residual measurement error of x1
    elseif kk==2
        x2_0 = x_k;                                         % the final estimate of x2 by separate recovery
        r2_k = r_k;                                         % residual measurement error of x2
    end
end

clear comp comp1 comp2 W1 W2 rootPSD_HT rootPSD_Wie rootPSD_Ae ... 
    x_k r_k z_k z_k_n u_k u_k_1 y A At S X w_k t_k alpha_k alpha_k_1

%% Automatic registration from the separately recovered images, Eq.(7)

% we consider the separately recovered x1_0 as the moving image, and x2_0 
% as the reference image for multiscale dense registration, and adopt 
% the co-registered image "x_0" for initialization of the joint recovery phase
visualize_reg     = 1;                                      % visualizing the coarse-to-fine registration
%%% COARSE-TO-FINE CO-REGISTRATION 
[x_0(:,:,1), x_0(:,:,2), affCoefs] = Regist(x1_0,x2_0,visualize_reg); 

x_org(:,:,1)      = Affine_Deform(x1, affTransf_x2);        % "x_org" : co-registered ground-truth, we register
                                                            % the moving ground-truth image by the oracle deformation parameters
x_org(:,:,2)      = x2;

%% Computing compensation arrays
% Computing the PSD of joint measurement error counterpart w.r.t. the sparsifying transforms used internally
% by the filter and w.r.t. the estimated deformation parameters used for co-registration

blkSize           = 8;                                      % for the joint recovery we employ BM4D 
                                                            % operating on cubes of size 8x8x2
HT_Tr_12          = 'dct';                                  % sparsifying block transform
HT_Tr_3           = 'dct';                                  % sparsifying transform along the third (cross-modal) dimension

L                 = 31;
rootPSD_Ad_e1     = Est_rootPSD(real(Ad_e1),[L L],...
      {'fft', 'fft'}, 0, 'sVar');                           % computing the FFT-rootPSD of the measurement error of the moving image
                                                            
                                                                           
rootPSD_Ad_e2     = Est_rootPSD(real(Ad_e2),[blkSize blkSize],...
      {HT_Tr_12, HT_Tr_12}, 0, 'sVar');                     % computing the rootPSD of the the measurement error
                                                            % counterpart w.r.t. the sparsifying basis used in BM4D (i.e., DCT)                    
                                                            
                                                            
% 2D compensation array w.r.t. FFT basis (for the moving system)
[Tr_12, ~]        = GetTransformMatrix(L, 'fft');
% % Tr_12             = fft(eye(L)); % equivalent
W1                = zeros(L,L);
for j1  = 1:L
    for j2        = 1:L
        Z         = zeros(L);
        Z(j1,j2)  = 1;
        B         = conj(Tr_12')*Z*Tr_12;
        W1(j1,j2) = sum(sum(abs((fft2(B,height1, width1)).*S1).^2))/(height1*width1)/L^2;
    end
end
compn_FFT         = 1./sqrt(W1);                            % FFT-root-PSD compensation array of A1^{\dagger}A1.

% estimating the deformation parameters of the reference system                                                                         
[~,~,~,est_defpar_x2] = Decompose_Affine(affCoefs, size(x1_0)); 
% computing the affine coeffients (using only rotation and scaling properties of Fourier transform) for deforming the kernel
Kernel_affCoefs   = Affine_Transf([0 0 est_defpar_x2(3:4)], size(x2));



% 2D compensation array w.r.t. DCT basis (for the reference system)
[Tr_12, ~]        = GetTransformMatrix(blkSize, HT_Tr_12);
% % Tr_12           = dct(eye(blkSize)); % equivalent
W2                = zeros(blkSize);               
for j1 = 1:blkSize
    for j2        = 1:blkSize
        Z         = zeros(blkSize);
        Z(j1,j2)  = 1;
        B         = Tr_12'*Z*Tr_12;
        W2(j1,j2) = sum(sum(abs((fft2(B,height2, width2)).*S2).^2))/(height2*width2);
    end
end
compn_factor2     = 1./sqrt(W2);                            % DCT root-PSD compensation array of A2^{\dagger}A2.


[Tr_3, ~]         = GetTransformMatrix(size(x_0,3), HT_Tr_3);  % the transform basis along the cross-modal dimension
% % Tr_3          = dct(eye(size(x_0,3)));  % equivalent


%% Joint reconstruction using FBS and BM4D as the denoiser -- Tackling Eq.(5)

PnP_solver_joint  = PnP_solver_sep;                         % PnP apprtoach for joint recovery phase
                                                            % (can be different than the separate one)
OutLoopIter       = 10;                                     % max iteration number
PSNR_Joint        = zeros(OutLoopIter,1);
lambda_bm4d       = 2.7;
alpha_k           = 1;
u_k               = x_0;
x_k               = x_0;

if experiment == 1  % multichannel sparse recovery
    % step size (can be different for each modality and PnP approach)
    % it is optimized so to obtain the highest PSNR at iteration k=10
    if strcmpi(PnP_solver_joint,'FISTA')
        Rho       = [1.2 1];
    elseif strcmpi(PnP_solver_joint,'FBS')
        Rho       = [1.2 1];
    elseif strcmpi(PnP_solver_joint,'ISTA')
        Rho       = [1.2 1];
    else
        error('Your solver does not exist!')
    end
    
else               % multicontrast sparse MR recovery
    % step size (can be different for each modality and PnP approach)
    % it is optimized so to obtain the highest PSNR at iteration k=10
    if strcmpi(PnP_solver_joint,'FISTA')
        Rho       = [.4 .2];
    elseif strcmpi(PnP_solver_joint,'FBS')
        Rho       = [.4 .2];
    elseif strcmpi(PnP_solver_joint,'ISTA')
        Rho       = [.4 .2]; 
    else
        error('Your solver does not exist!')
    end
end


Rho          = cat(3,Rho(1),Rho(2));                        % the joint step-size

for k = 1:OutLoopIter  %%% iterative joint recovery %%%
     
    R1 = real(At1(r1_k));                                   % the residual image of x_1      
    R2 = real(At2(r2_k));                                   % the residual image of x_2
    
    %%% estimating the 8x8 DCT-root-PSD of the effective noise in the co-registered moving image at ieration k
    % computing LxL FFT-root-PSD of residual measurement image R1
    FFT_rPSD_R1   = Est_rootPSD(R1, [L L], {'fft','fft'}, 0, 'sVar'); 
    % computing LxL FFT-PSD of effective noise in the moving image
    FFT_PSD_eta1  = ( Rho(1).*FFT_rPSD_R1 + abs(FFT_rPSD_R1-rootPSD_Ad_e1).*compn_FFT ).^2;
    % computing global FFT-PSD of the effective noise in the moving image
    Fourier_PSD_eta1 = FFT2Fourier_PSD_convertor(FFT_PSD_eta1, [height1, width1]);
    % the convolution kernel corresponding to the global FFT-PSD "Fourier_PSD_eta1"
    Kernel_eta1   = ifftshift(ifft2(sqrt(Fourier_PSD_eta1/numel(Fourier_PSD_eta1))));
    % deforming the kernel w.r.t. the scale and rotation parameter estimated in the co-registration step
    deform_Kernel = Affine_Deform(Kernel_eta1,Kernel_affCoefs)./est_defpar_x2(3);
    % computing the global FFT-PSD w.r.t. deformed kernel (i.e., corresponding to the global FFT-PSD
    % of the effective noise in the registered moving image)
    deform_Fourier_PSD = abs(fft2(deform_Kernel, height1, width1)).^2*(height1*width1);
    % converting the global FFT-PSD into the DCT root-PSD 
    rPSD_eta1     = sqrt( PSD_domain_convertor(deform_Fourier_PSD, [blkSize blkSize], {HT_Tr_12, HT_Tr_12}) );
  
    
    %%% estimating the 8x8 DCT-root-PSD of the effective noise in the reference image at ieration k
    % computing 8x8 DCT-root-PSD of residual measurement image R2
    rPSD_R2       = Est_rootPSD(R2, [blkSize blkSize], {HT_Tr_12, HT_Tr_12}, 0, 'sVar');
    % computing 8x8 DCT-root-PSD of the effective noise in the reference image
    rPSD_eta2     = Rho(2).*rPSD_R2 + abs(rPSD_R2 - rootPSD_Ad_e2).*compn_factor2;
    
    %%% computing the joint effective noise root-PSD
    rootPSD       = sqrt( reshape([rPSD_eta1(:).^2,rPSD_eta2(:).^2]*(Tr_3').^2,...
                          [blkSize blkSize size(x_0,3)]) );              
    rootPSD       = double(rootPSD); 
    

    Res(:,:,1)    = Affine_Deform(R1, affCoefs);            % co-registering the residual images
    Res(:,:,2)    = R2;   

    z_k           = x_k + Rho.*Res;                         % noisy signal to be filtered
    
    u_k_1         = u_k;
    
    %%% filtering by stationary correlated noise denoiser
    u_k           = bm4d_colored(z_k, 1, lambda_bm4d, rootPSD);  % BM4D filtering
    
    alpha_k_1     = alpha_k;
    alpha_k       = .5+.5*sqrt(1+4*alpha_k^2);
    
    if strcmpi(PnP_solver_joint,'FISTA')
        t_k       = (alpha_k_1-1)/alpha_k;
    elseif strcmpi(PnP_solver_joint,'FBS')
        if k<160, t_k   = (alpha_k^k-1)/(alpha_k^(k+1)); else, t_k = 0; end
    elseif strcmpi(PnP_solver_joint,'ISTA')
        t_k       = 0;
    end
    x_k           = u_k + t_k.*(u_k-u_k_1);                 % updating x_k
 
    
    %%% mapping the jointly recovered images into their original grid (i.e., mis-registering
    %%% them in order to compute their measurement residual error at iteration k)
    [X1_k, ind]   = Affine_Deform(x_k(:,:,1), affCoefs, 'backward'); 
    X1_k(ind==0)  = x1_0(ind==0);
     X2_k         = x_k(:,:,2);
  
     r1_k         = y1 - A1(X1_k);                          % computing measurement residual error #1
     r2_k         = y2 - A2(X2_k);                          % computing measurement residual error #2

    PSNR_Joint(k) = 10*log10((max(x_org(:))).^2/mean((x_org(:)-x_k(:)).^2));
    fprintf('Joint Recovery, Iteration = %d, PSNR = %0.3f dB\n',...
               k, PSNR_Joint(k));
    
end

%% Visualization

if experiment ==1
    SC      = @(w1,w2) max(min(w1,max(w2(:))),min(w2(:)));  % for intensity clipping
    gamma   = .8;                                           % gamma correction factor
    C_org   = imfuse(x_org(:,:,1).^gamma,x_org(:,:,2).^gamma,'falsecolor','Scaling','none','ColorChannels',[2 1 2]);
    C_sep   = imfuse(SC(x_0(:,:,1),x_org(:,:,1)).^gamma,SC(x_0(:,:,2),x_org(:,:,2)).^gamma,...
        'falsecolor','Scaling','none','ColorChannels',[2 1 2]);
    C_joint = imfuse(SC(x_k(:,:,1),x_org(:,:,1)).^gamma,SC(x_k(:,:,2),x_org(:,:,2)).^gamma,...
        'falsecolor','Scaling','none','ColorChannels',[2 1 2]);

    images = insertText([C_org,C_sep,C_joint],[410 10; 952 10; 1492 10]...
        ,{'ground-truth','separate','joint'},'FontSize',18,'BoxColor',...
       {'red','red','red'},'BoxOpacity',0.4,'TextColor','white');
    figure, imshow(images)
    
else
    SC      = @(w1,w2) max(min(w1,max(w2(:))),min(w2(:)));  % for intensity clipping
    gamma   = .8;                                           % gamma correction factor
    x_sep(:,:,1)   = SC(x_0(:,:,1),x_org(:,:,1));
    x_sep(:,:,2)   = SC(x_0(:,:,2),x_org(:,:,2));
    x_joint(:,:,1) = SC(x_k(:,:,1),x_org(:,:,1));
    x_joint(:,:,2) = SC(x_k(:,:,2),x_org(:,:,2));
    
    for i=1:2
        figure, imshow([x_org(:,:,i),x_sep(:,:,i),x_joint(:,:,i)].^gamma,[]);
        text(size(x_org,1)*0.5,-1.5,['ground-trouth T' num2str(i)],'Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom');
        text(size(x_org,1)*1.25,-1.5,'separately recovered','Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom');
        text(size(x_org,1)*2.1,-1.5,'jointly recovered','Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom'); axis off, axis tight
    end
end


PSNR_separate = 10*log10((max(x_org(:))).^2/mean((x_org(:)-x_0(:)).^2)); 
PSNR_joint = PSNR_Joint(end); 

fprintf('-------------------------------------------------------\n')
fprintf('separate sparse recovery: PSNR = %0.3f dB\n',PSNR_separate); 
fprintf('   joint sparse recovery: PSNR = %0.3f dB\n',PSNR_joint); 
