function [x1_def, x2_def, affCoefs] = Regist(x1,x2, visualize) 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Regist performs multimodal coarse-to-fine block-wise image registration.
%
%  N. Eslahi and A. Foi, "Joint sparse recovery of misaligned multimodal images
%  via daptive local and nonlocal cross-modal", in Proc. IEEE CAMSAP (Dec. 2019) 
%
%
%  FUNCTION INTERFACE:
%  [x1_def, x2_def, affCoefs] = Regist(x1,x2, visualize)
% 
%  INPUTS:
%     1) x1             (2D array)  : the moving images
%     2) x2             (2D array)  : the reference image
%     3) visualize      (logical)   : 0 --> do not visualize output information
%                                   : 1 --> visualize information to screen
%                                         (default is 0)
%
%  OUTPUTS:
%     1) x1_def          (2D array) : the registered (deformed) moving image
%     2) x2_def          (2D array) : the reference (fixed) image
%     3) affCoefs        (2x3 array): the affine deformation/transformation matrix
%                                         
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
% AUTHOR:
%                     Alessandro Foi & Nasser Eslahi
%     email:           firstname.lastname@tuni.fi
%            
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~exist('visualize','var')||isempty(visualize)
    visualize = 0;
end

showEveryNseconds=.5;
affCoefs   = [0 0;1 0;0 1]; % initialization with no deformation

%%% BUILDS DERIVATIVE KERNELS

L          = 5; % window has size LxL
[t_2, t_1] = meshgrid(1:L,1:L); % indep. variables (horiz. and vert.)
t_1        = t_1-mean(t_1(:)); % define t_1 and t_2 so that they are centered and with unit step
t_2        = t_2-mean(t_2(:)); % ...
N          = 2; % polynomial order (bivariate)
blkM       = zeros(L^2,(N+1)*(N+1-N/2));
k_table    = zeros((N+1)*(N+1-N/2),2);
ctr = 0;
for j=0:N
    for k_1 = 0:j % degree on t_1
        k_2 = j-k_1; % degree on t_2
        ctr = ctr+1;
        blkM(:,ctr)    = 1/factorial(k_1+k_2)*(t_1(:).^k_1.*t_2(:).^k_2); % monomials like in MacLaurin        
        k_table(ctr,:) = [k_1 k_2]; % table with powers
    end
end
s          = L/5; % relative spread of Gaussian window (as stdev of Gaussian)
w          = exp(-((0.5/s^2)*t_2.^2+(0.5/s^2)*t_1.^2)); % Gaussian window
blkMtilde  = blkM*pinv(blkM'*diag(w(:))*blkM); % THE DUAL FRAME
der_order  = [0 1]; % desired order of partial derivatives on t_1 and t_2
blkDiff    = false(1,size(blkMtilde,2));  % indicator vector that finds the position of the ...
blkDiff(all(k_table==der_order,2))=1; % ... corresponding monomial in the frame
K          = reshape((blkMtilde(:,blkDiff).*w(:))',[L L]); % convolut. kernel for differentiation of order der_order
K          = K(end:-1:1,end:-1:1); % flipping of the kernel (a.k.a. mirroring)



%%%analysis with dual frame (convolut. kernel)

H0         = repmat(((1:size(x2,2)) -0.5),[size(x2,1) 1]);  % horz. coord. of pixels within full-size image
V0         = repmat(((1:size(x2,1))'-0.5),[1 size(x2,2)]);  % vert. coord. of pixels within full-size image

smlSzs     = round(exp(linspace(log(25),log(max(size(x2,1),size(x2,2))+30),2*300+445)));
smlSzs     = min(size(x2,1),max(1,smlSzs));
smlSzs     = [ones(1,10)*smlSzs(1) smlSzs ones(1,5)*smlSzs(end)];
iterCount  = 0;
timeOld    = -inf;
timeNow    = now;
for smlSz1 = smlSzs  %% start iterative registration
    iterCount = iterCount+1;
    mu     = 0.6; % damping coefficient
    alpha  = 1; 
    
    smlSz2 = smlSz1;  % size of extracted blocks
    pxSz   = size(x2,1)/smlSz1;  % relative pixel size of small images
    
    
    % blockwise processing
    blkSz  = min(31,round(13*size(x2,1)/40/pxSz));
    blkStp = max(1,round(0.2*4*512/40/pxSz));
    blkJ1  = unique([1:blkStp:smlSz1-blkSz+1  smlSz1-blkSz+1]); % vert. indices of blocks
    blkJ2  = unique([1:blkStp:smlSz2-blkSz+1  smlSz2-blkSz+1]); % horz. indices of blocks
    
    blkH0  = repmat(pxSz*(blkJ2 +(blkSz-2)/2),[numel(blkJ1) 1]); % horz. coord. of block centers
    blkV0  = repmat(pxSz*(blkJ1'+(blkSz-2)/2),[1 numel(blkJ2)]); % vert. coord. of block centers
    blkH0  = blkH0(:);
    blkV0  = blkV0(:);
    
    slmH0  = repmat(pxSz*((1:smlSz2) -0.5),[smlSz1 1]);  % horz. coord. of pixels within small image
    smlV0  = repmat(pxSz*((1:smlSz1)'-0.5),[1 smlSz2]);  % vert. coord. of pixels within small image
    slmH0  = slmH0(:);
    smlV0  = smlV0(:);
    
    
    % interpolators for x1 and x2
    absDet = abs(det(affCoefs(2:3,:)));
    kerStd = pxSz*0.75;
    smoothKer1 = fspecial('gaussian',[round(kerStd*2)*2+1 1],kerStd*sqrt(absDet));
    smoothKer2 = fspecial('gaussian',[round(kerStd*2)*2+1 1],kerStd);
    smoothMe   = @(x,k) padarray(conv2(x,k,'valid'),(size(k)-1)/2,'both','replicate');
    smoothMeTwice = @(x,k) smoothMe(smoothMe(x,k),k');
    
    if iterCount>round(.75*numel(smlSzs))
        interpMethod = 'spline';
        F  = griddedInterpolant(V0,H0,smoothMeTwice(x1,smoothKer1),interpMethod,'nearest');
        F2 = griddedInterpolant(V0,H0,smoothMeTwice(x2,smoothKer2),interpMethod,'nearest');
    else
        interpMethod = 'linear';
        F  = griddedInterpolant(V0,H0,smoothMeTwice(x1,smoothKer1),interpMethod);
        F2 = griddedInterpolant(V0,H0,smoothMeTwice(x2,smoothKer2),interpMethod);
    end

    % goes to small size
    x2_sml = reshape(F2(smlV0,slmH0),[smlSz1 smlSz2]);
    
    smlM   = [ones(numel(smlV0),1) slmH0 smlV0]; % set of vectors (lin. indep. unless we are very unlucky or n>d)
    smlHV  = smlM*affCoefs; % synthesis with frame and coefficients from inner product with the dual frame
    x1_sml = reshape(F(smlHV(:,2),smlHV(:,1)),[smlSz1 smlSz2]);
    
    
    
    % computes gradient of x2_temp (by convolution against kernels)
    grad_x2_sml = cat(3,padarray(conv2(x2_sml,K ,'valid'),(size(K )-1)/2,'replicate'), ...
        padarray(conv2(x2_sml,K','valid'),(size(K')-1)/2,'replicate'));
    
    % compute blockwise shift vector field
    shftVecFld  = zeros(numel(blkJ1)*numel(blkJ2),2);  %% buffer for shift vectors for blocks
    E      = zeros(numel(blkJ1)*numel(blkJ2),1);   %% buffer for error of prediction based on shift
    Ge     = zeros(numel(blkJ1)*numel(blkJ2),1);   %% buffer for gradient as contrast estimate
    
    c12    = 0;  %% counter for both H&V
    for j2 = blkJ2
        for j1 = blkJ1
            c12       = c12+1;
            
            blk1      = x1_sml(j1:j1+blkSz-1,j2:j2+blkSz-1); % extracted block from coarse x1
            blk2      = x2_sml(j1:j1+blkSz-1,j2:j2+blkSz-1); % extracted block from coarse x2
            
            grad_blk2 = grad_x2_sml(j1:j1+blkSz-1,j2:j2+blkSz-1,:); % extracted block from the gradient of coarse x2
            grad_blk2 = reshape(grad_blk2,blkSz*blkSz,2);
            
            M         = blk1(:).^(0:3);
            blk1      = M*(M\blk2(:));  % faster than blk1=M*pinv(M)*blk2(:); 
                                        % maps intensities of blk1 to those of blk2 
                                        % (preferred, as we use x2 as better-quality reference also for gradient values)

            blkDiff   = blk2(:)-blk1(:);    % difference (after intensity mapping, treat blk1 as shifted version of blk2)
            
            shftVec   = blkDiff\grad_blk2;  % estimates the shift vector (least-squares fit of differences by gradient)
            shftVecFld(c12,:) = shftVec;  % building shift vector fied as collection of shift vectors
            
            e         = eps+mean((blkDiff-grad_blk2*shftVec').^2);  % mean quadratic error
            E(c12)    = e;
            Ge(c12)   = 2*mean(grad_blk2(:).^2);  % local contrast estimate as mean of gradient vector squared magnitude (square length)
        end
    end
    shftVecFld   = pxSz*shftVecFld;  %% scales vectors by relative pixel size

    % here we fit a global affine transformation to the blockwise shift vector field
    blkM         = [ones(numel(blkV0),1) blkH0 blkV0]; % set of vectors (lin. indep. unless we are very unlucky or n>d) %% the ones in the first column represents a unit shift
    w            = (1./sqrt(E(:)./Ge(:))); w(isnan(w))=0; w(isinf(w))=0; % non-negative, weights (can be zero)
    blkMtilde    = blkM*pinv(blkM'*diag(w)*blkM); % THE DUAL FRAME
    affCoefs_new =(blkMtilde'.*w')*((alpha*[zeros(numel(blkV0),1) shftVecFld]+blkM)*affCoefs); % coefficients from inner product with the dual frame
    % updating the affine coefficient (i.e. affine deformation matrix)
    affCoefs     = affCoefs*mu + (1-mu)*affCoefs_new;  %% damping
    
    smlHV        = smlM*affCoefs; % synthesis with frame and coefficients from inner product with the dual frame
    x1_sml       = reshape(F(smlHV(:,2),smlHV(:,1)),[smlSz1 smlSz2]);

    warning off
        
        if visualize
            
            disp(' ');
            disp(['# ',num2str(iterCount),'  smlSz: ',num2str(smlSz1)])
            disp(num2str(affCoefs))
            
            if ((timeNow-timeOld)*86400)>showEveryNseconds
                blkHV   = blkM*affCoefs; % synthesis with frame and coefficients from inner product with the dual frame
                blkH    = blkHV(:,1);
                blkV    = blkHV(:,2);
                x1_sml0 = reshape(F(smlV0,slmH0),[smlSz1 smlSz2]);
                subplot(2,1,1)
                imagesc(max(0,min(1,...
                    [x1_sml0./max(x1_sml0(:)),x1_sml./max(x1_sml(:)),x2_sml./max(x2_sml(:))]))),axis equal tight
                subplot(2,2,3)
                quiver(blkH,blkV,blkH0-blkH,blkV0-blkV,0);
                hold on
                quiver(blkH0,blkV0,reshape(shftVecFld(:,1),size(blkH)),reshape(shftVecFld(:,2),size(blkV)),0);
                hold off
                set(gca,'YDir','reverse');
                axis equal;
                set(gca,'xlim',[0 size(x2,2)],'ylim',[0 size(x2,1)]);
                drawnow
                title(['# ',num2str(iterCount),'  smlSz: ',num2str(smlSz1),'  blkSz: ',num2str(blkSz)]);
                pause(0.00001)
                timeOld = timeNow;
            end
            timeNow = now;
 
        end
        
        

end

x2_def = x2;

F      = griddedInterpolant(V0,H0,x1,'spline');
x1_def = reshape(F([ones(numel(x1),1) H0(:) V0(:)]*affCoefs(:,2),[ones(numel(x1),1) H0(:) V0(:)]*affCoefs(:,1)),size(x2));

if visualize  
   
subplot(2,2,4)

hhh    = imagesc(x1_def);set(hhh,'AlphaData',0.50);
hold on
hhh0   = imagesc(x1);set(hhh0,'AlphaData',0.25)

blkHV  = blkM*affCoefs; % synthesis with frame and coefficients from inner product with the dual frame
blkH   = blkHV(:,1);
blkV   = blkHV(:,2);

hhhq   = quiver(blkH,blkV,blkH0-blkH,blkV0-blkV,0);

set(hhhq,'LineWidth',0.5,'Color',[0 0 0])
axis equal tight
hold off
 
end


