function x_noisy = Additive_WGN(x,snr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Additive_WGN Adds zero-mean real-/complex-valued white Gaussian noise to
% the signal "x" yielding signal-to-noise ratio of "snr" dB with respect to 
% the power of signal, equivalent to:
%
%               x_noisy = awgn(x,snr,'measured')
%
% (useful when "Communication System Toolbox" is not installed in your MATLAB.)
% 
%  FUNCTION INTERFACE:
%  x_noisy = Additive_WGN(x,snr)
%
%  INPUTS:
%     1) x               :Input signal
%     2) snr             :Signal-to-noise ratio (in dB)
%
%
%  OUTPUTS:
%     1) x_noisy         :Noisy signal
%
%-------------------------------------------------------------------------
%  EXAMPLES:
%
% x   = im2double(imread('cameraman.tif'));
% snr = 10;
% rng(0,'v4')
% x_noisy1 = awgn(x,snr,'measured');
%
% rng(0,'v4')
% x_noisy = Additive_WGN(x,snr);
%
% diff = x_noisy1 - x_noisy;
% [max(diff(:)) min(diff(:))]
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if isreal(x)==1
        WGN      = randn(size(x));
        std_WGN  = sqrt(mean(x(:).^2)./(10.^(snr/10)));
        x_noisy  = x + std_WGN .* WGN;
    else
        real_WGN = randn(size(x));
        imag_WGN = randn(size(x));
        std_WGN  = sqrt(mean(abs(x(:)).^2)./(10.^(snr/10)));
        x_noisy  = x + (std_WGN/sqrt(2)) .* (real_WGN + 1i.* imag_WGN);
    end
end



