function [x_def, indicatorMap, final_grid] = Affine_Deform(x,affCoefs,action)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Affine_Deform deforms the input image according to the deformation
% matrix "affCoefs"
%
%  FUNCTION INTERFACE:
%   [x_def, indicatorMap, final_grid] = Affine_Deform(x,affCoefs,action)
%
% INPUTS:
%   x:                (double)          Input 2D image.
%
%   affCoefs:         (3x2 array)       affine deformation matrix,
%                                       for more information see the syntaxes
%                                       of "Affine_Transf.m", "decomposeAffine.m", 
%
%   action:           (string)          'forward' or 'backward' projection w.r.t. the
%                                       deformation matrix "affCoefs".
%                                       'forward' implements the deformation of "x" w.r.t. the 
%                                       "affCoefs", while 'backward' implements the inverse 
%                                       deformation of "x" w.r.t. the "affCoefs".
%
% OUTPUTS:
%   z:                (double)          deformed image.
%
%   indicatorMap:     (array/logic)     binary map indicating positions of the pixels
%                                       from old grid into the new (deformed) grid
%
%   final_grid        (complex double)  the new (deformed) grid
%
%
%  EXAMPLES:
%-------------------------------------------------------------------------
%      horizontal_shift = 5.3;    % to the right side
%      vertical_shift   = -2.8;   % to the top side
%      zoom             = 1.25;   % >1 : zoom-in, <1 : zoom-out
%      rotation         = 10;     % >0 counterclockwise rotation (in degree),
%                                 % <0 clockwise rotation
%      shear            = [0 .1]; % shearing parameter, one of the should
%                                 % be zeros
%      deformParameters = [horizontal_shift, vertical_shift, zoom, rotation, shear];
% 
%      x                = im2double(imread('cameraman.tif')); 
% 
%      affCoefs         = Affine_Transf(deformParameters,size(x)); % computing the affine deformation matrix 
%                                                                  % w.r.t. the deformation parameters
%      [x_def, forwardMap, fgrid]  = Affine_Deform(x,affCoefs);    % forward deformation with Affine_Deform
% 
%      [x_hat, backwardMap, bgrid] = Affine_Deform(x_def,affCoefs,'backward'); % backward deformation with 
%                                                                              % Affine_Deform
% 
%      figure, subplot(1,3,1), imshow(x), title([{'original img.'}, ... 
%                  '$\mathbf{x}$'],'Interpreter','Latex') 
%              subplot(1,3,2), imshow(x_def), title([{'deformed img.'}, ...
%                  '$\mathbf{z}=\mathcal{D}(\mathbf{x},\boldmath{\theta})$'],'Interpreter','Latex') 
%              subplot(1,3,3), imshow(x_hat), title([{'inv. deformed img.'}, ...
%                  '$\hat{\mathbf{x}}=\mathcal{D}^{-1}(\mathbf{z},\boldmath{\theta})$'],'Interpreter','Latex') 
% 
%-------------------------------------------------------------------------
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
% AUTHORS:
%     Diego Stucchi, Nasser Eslahi, Alessandro Foi
%     email:       stucchidiego1994@gmail.com
%            {nasser.eslahi, alessandro.foi}@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~exist('action','var')||isempty(action)
    action      = 'forward';
end

finalSize    = size(x);

if strcmpi(action, 'forward') == 1
    g1         = repmat(((1:size(x,2)) -0.5),[size(x,1) 1]);
    g2         = repmat(((1:size(x,1))'-0.5),[1 size(x,2)]);
        
    final_grid = [ones(numel(g2),1) g1(:) g2(:)];
    final_grid = final_grid*affCoefs;

    F = griddedInterpolant(g2,g1,x,'spline','nearest');
% F = griddedInterpolant(g2,g1,x,'linear','nearest');
    x_def = reshape(F(final_grid(:,2),final_grid(:,1)),finalSize);
    
    indicatorMap  = final_grid(:,1)<=size(x,2) & final_grid(:,2)<=size(x,1) ...
        & final_grid(:,1)>=1 & final_grid(:,2)>=1;
    indicatorMap  = reshape(indicatorMap,finalSize);
    
    
elseif strcmpi(action, 'backward') == 1
    g1         = repmat(((1:size(x,2)) -0.5),[size(x,1) 1]);
    g2         = repmat(((1:size(x,1))'-0.5),[1 size(x,2)]);
    
    final_grid    = [g1(:) g2(:)] - affCoefs(1,:);
    final_grid    = final_grid / (affCoefs(2:3,:));
    
    indicatorMap  = final_grid(:,1)<=size(x,2) & final_grid(:,2)<=size(x,1) ...
        & final_grid(:,1)>=1 & final_grid(:,2)>=1;
    indicatorMap  = reshape(indicatorMap,finalSize);
    
    F = griddedInterpolant(g2,g1,x,'spline','nearest');
% F = griddedInterpolant(g2,g1,x,'linear','nearest');
    x_def = reshape(F(final_grid(:,2),final_grid(:,1)),finalSize);

else
    error('acceptable actions are: ''forward'' and ''backward''')
end

final_grid = final_grid(:,1) + 1i * final_grid(:,2);
final_grid = reshape(final_grid,size(x_def));
