function affCoefs = Affine_Transf(deformParameters,imgSize)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Affine_Transf computes the affine transformation matrix corresponding to the
% given affine deformation parameters, i.e. deformParameters = [horizontal_shift, ...
% vertical_shift, zoom, rotation, horizontal_Shear, vertical_Shear]
%
% We assume that the coordinates grid starts from 0.5 in both horizontal and 
% vertical dimensions
%
%  FUNCTION INTERFACE:
%   affCoefs = Affine_Transf(deformParameters,imgSize)
%
% INPUTS:
%   deformParameters: (1x4 or 1x6 array) Array with deformation parameters:
%                                        [horizontal_shift vertical_shift zoom rotation...
%                                          horizontal_Shear vertical_Shear], s.t.
%                                        (horizontal_Shear * vertical_Shear = 0, 
%                                        i.e. shear in only one of the two directions)
%   imagSize:         (1x2 array)        the size of image to be deformed
%
% OUTPUTS:
%   affCoefs:         (3x2 array)        affine deformation matrix
%
%  EXAMPLES:
%-------------------------------------------------------------------------
%      horizontal_shift = 5.3;    % horizontal shift to the right side
%      vertical_shift   = -2.8;   % vertical shift to the top side
%      zoom             = 1.25;   % >1 : zoom-in, <1 : zoom-out
%      rotation         = 10;     % >0 counterclockwise rotation (in degree),
%                                 % <0 clockwise rotation
%      shear            = [0 .1]; % shearing parameter, one of the should be zeros
%      deformParameters = [horizontal_shift, vertical_shift, zoom, rotation, shear];
%      imgSize          = [512 512]; % size of the image to be deformed
%      % computing affine deformation matrix
%      affCoefs         = Affine_Transf(deformParameters, imgSize);
%      % computing the deformation parameters from the deformation matrix
%      [~,~,~,deformParameters_hat] = Decompose_Affine(affCoefs, imgSize);
%-------------------------------------------------------------------------
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
% AUTHORS:
%     Diego Stucchi, Nasser Eslahi, Alessandro Foi
%     email:       stucchidiego1994@gmail.com
%            {nasser.eslahi, alessandro.foi}@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pars = size(deformParameters,2);
if pars ~= 4 && pars ~= 6
    affCoefs = NaN;
    error('Error in deformParameters! deformParameters = [hShift, vShift, zoom, rotAngle, xShear, yShear]');
elseif pars == 4
    % No shearing
    deformParameters(:,5:6) = 0;
end

% THIS WILL BE REMOVED WHEN MIXED SHEARING WILL BE IMPLEMENTED
tmp = deformParameters(:,5) .* deformParameters(:,6);
if sum(tmp) ~= 0
    affCoefs = NaN;
    error('Mized shearing not implemented yet!');
end
    

% The affCoefs will be stacked in the 3rd dimension
n = size(deformParameters,1);
affCoefs = zeros(3,2,n);
for k = 1 : n

    hShift = deformParameters(k,1);
    vShift = deformParameters(k,2);
    zoom   = deformParameters(k,3);
    theta  = (pi * deformParameters(k,4)) / 180;
    bx     = deformParameters(k,5);
    by     = deformParameters(k,6);
    
    ctr1 = imgSize(2)/2;
    ctr2 = imgSize(1)/2;
    
    S = [1 bx;by 1];
    K = [1/zoom, 0; 0, 1/zoom];
    R = [cos(theta),-sin(theta);sin(theta),cos(theta)];
    
    q = [ctr1;ctr2] - S*K*R*[ctr1;ctr2] - [hShift;vShift];
    A = S*K*R;
    
    affCoefs(:,:,k) = [q';A'];

end


end

