function [S,K,R,deformParameters] = Decompose_Affine(affCoefs, imgSize)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Decompose_Affine estimate the deformation parameters, i.e. 
% deformParameters = [shiftH, shiftV, zoom, rotationAngle, shearX, shearY],
% from the affine deformation matrix.
%
%  FUNCTION INTERFACE:
%   [S,K,R,deformParameters] = Decompose_Affine(affCoefs,imgSize)
%
% INPUTS:
%   affCoefs:         (3x2 array)        affine deformation matrix
%   imagSize:         (1x2 array)        the size of image to be deformed
%
% OUTPUTS:
%   S:                (2x2 array)        shearning matrix
%   K:                (2x2 array)        scaling matrix
%   R:                (2x2 array)        rotation matrix 
%   deformParameters  (1x6 array)        [horizontal_shift vertical_shift zoom rotation...
%                                         horizontal_Shear vertical_Shear], s.t.
%                                        (horizontal_Shear * vertical_Shear = 0, 
%                                        i.e. shear in only one of the two directions)
%
%  EXAMPLES:
%-------------------------------------------------------------------------
%      horizontal_shift = 5.3;    % horizontal shift to the right side
%      vertical_shift   = -2.8;   % vertical shift to the top side
%      zoom             = 1.25;   % >1 : zoom-in, <1 : zoom-out
%      rotation         = 10;     % >0 counterclockwise rotation (in degree),
%                                 % <0 clockwise rotation
%      shear            = [0 .1]; % shearing parameter, one of the should be zeros
%      deformParameters = [horizontal_shift, vertical_shift, zoom, rotation, shear];
%      imgSize          = [512 512]; % size of the image to be deformed
%      % computing affine deformation matrix
%      affCoefs         = Affine_Transf(deformParameters, imgSize);
%      % computing the deformation parameters from the deformation matrix
%      [~,~,~,deformParameters_hat] = Decompose_Affine(affCoefs, imgSize);
%-------------------------------------------------------------------------
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
% AUTHORS:
%     Diego Stucchi, Nasser Eslahi, Alessandro Foi
%     email:       stucchidiego1994@gmail.com
%            {nasser.eslahi, alessandro.foi}@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

A = affCoefs(2:3,:)';
q = affCoefs(1,:)';

deformParameters = zeros(1,6);

if ~exist('q','var')
    deformParameters(1:2) = [0 0];
else
    if ~exist('imgSize','var')
        error('The image size is required in order to reconstruct the shift parameters');
    else
        % Translation
        ctr = [imgSize(2) imgSize(1)] / 2;
        v = (eye(2) - A) * ctr' - q;
        deformParameters(1:2) = v;
    end
end

M = A*A';
% Scaling
h = nthroot(det(M),4);
K = h * eye(2);
deformParameters(3) = 1/h;

M = M / (h*h);

A = A / h;
% Rotation
if abs(M(1,1)-1) <= abs(M(2,2)-1)
    deformParameters(5) = 0;
    deformParameters(6) = M(1,2);
    S = [1 0;M(1,2),1];
    t = atan(-A(1,2)/A(1,1));
else
    deformParameters(5) = M(1,2);
    deformParameters(6) = 0;
    S = [1 M(1,2);0,1];
    t = atan(A(2,1)/A(2,2));
end

R = [cos(t),-sin(t);sin(t),cos(t)];
deformParameters(4) = (180/pi) * t;

