function PSD_Fourier_domain = FFT2Fourier_PSD_convertor(PSD_FFT_domain,N)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  FFT2Fourier_PSD_convertor converts the input FFT-PSD ("PSD_FFT_domain") into
%  a global Fourier-domain PSD ("PSD_FFT_domain", interpolated FFT-PSD) of size "N".
%  
%
%  FUNCTION INTERFACE:
%     PSD_Fourier_domain = FFT2Fourier_PSD_convertor(PSD_FFT_domain,N)
% 
%  INPUTS:
%     1) PSD_FFT_domain       (2D/3D array) : FFT-PSD to be converted
%     2) N                    (double)      : the size of output PSD
%
%  OUTPUTS:
%     1) PSD_Fourier_domain   (2D/3D array) : converted Fourier-domain PSD
%
%
%  EXAMPLE:
%  prerequisite : Est_rootPSD.m (to compute the FFT-PSD using the sample variance)
%-------------------------------------------------------------------------
% N                    = [1023 511];                 % N1xN2: size of 2-D noise signal
% L                    = 11;                         % LxL: size of converted 2-D PSD
% w                    = randn(N);                   % standard white Gaussian noise of size N
% g1                   = fspecial('gaussian',7,1.5); % lowpass Gaussian filter 
% g2                   = fspecial('gaussian',7,0.5); % lowpass Gaussian filter
% g                    = g1-g2;                      % correlation kernel 
%                                                    % (bandpass Gaussian filter)
% eta                  = imfilter(w,g,'circular');   % correlated noise
%
% Fourier_PSD_kernel   = ...
%             abs(fft2(g,N(1),N(2))).^2*(numel(N));  % Fourier domain PSD of "eta"
%                                                    % computed from the kernel
% PSD_FFT_domain       = (Est_rootPSD(eta, [L L],...
%             {'fft','fft'}, 0, 'sVar')).^2;         % FFT-PSD of "eta"
%
% PSD_Fourier_domain = FFT2Fourier_PSD_convertor(PSD_FFT_domain,N);
%                                                    % converted PSD into Fourier domain
%                                                    % N can be diferent than the size of "eta"
% 
% figure, subplot(1,2,1), imagesc(Fourier_PSD_kernel), colormap('jet'),
% colorbar, title('the Fourier PSD computed from kernel')
%         subplot(1,2,2), imagesc(PSD_Fourier_domain), colormap('jet'),
% colorbar, title('the converted Fourier-domain PSD')
%-------------------------------------------------------------------------
%
%               AUTHORS:  Alessandro Foi & Nasser Eslahi
%                 contact: firstname.lastname@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

N1 = N(1);
N2 = N(2);
if numel(N)==2
    N3 = 1;
elseif numel(N)==3
    N3 = N(3);
else
    error('this version only accepts 2D and 3D PSDs!')
end


n1=size(PSD_FFT_domain,1);
B1=zeros(N1,n1);
for j1=1:n1
    eee=zeros(n1,1); eee(j1)=1;
    B1(1:n1,j1)=ifft(eee); 
    
end
Phi2_1=abs(fft(B1)).^2/N1^2;  
Phi2dual_1=Phi2_1*pinv(Phi2_1'*Phi2_1);    % dual frame
clear B1


n2=size(PSD_FFT_domain,2);
B2=zeros(N2,n2);
for j2=1:n2
    eee=zeros(n2,1); eee(j2)=1;
    B2(1:n2,j2)=ifft(eee);  
end
Phi2_2=abs(fft(B2)).^2/N2^2;  
Phi2dual_2=Phi2_2*pinv(Phi2_2'*Phi2_2);    % dual frame
clear B2



n3=size(PSD_FFT_domain,3);
B3=zeros(N3,n3);
for j3=1:n3
    eee=zeros(n3,1); eee(j3)=1;
    B3(1:n3,j3)=ifft(eee);  
end
Phi1_3=abs(fft(B3)).^2/N3^2;  
Phi1dual_3=Phi1_3*pinv(Phi1_3'*Phi1_3);    % dual frame
clear B3




% keyboard
PSD_Fourier_domain=zeros(N1,N2,N3);
cnvt_residual = PSD_Fourier_domain;
PSD_Fourier_2d = zeros(n1,n2,N3);
PSD_FFT_reshape = reshape(PSD_FFT_domain,[n1*n2 n3]);

Niter=100;
% keyboard
for jj=1:Niter
       
    rho_k=1;
    % 1) residual and update
    for i=1:N3
        PSD_Fourier_2d(:,:,i)=Phi2_1'*PSD_Fourier_domain(:,:,i)*Phi2_2;
    end
    PSD_Fourier_2d = reshape(PSD_Fourier_2d,[n1*n2 N3]);

    cnvt_Fourier_PSD = PSD_Fourier_2d*Phi1_3;
    clear PSD_Fourier_2d
    
    residual = PSD_FFT_reshape - cnvt_Fourier_PSD;
    
    cnvt_1d_residual = reshape( residual*Phi1dual_3', [n1 n2 N3]);
    for i=1:N3
        cnvt_residual(:,:,i)=Phi2dual_1*cnvt_1d_residual(:,:,i)*Phi2dual_2';
    end
    
    PSD_Fourier_domain = PSD_Fourier_domain + rho_k* cnvt_residual;
    
    % 2) constraints
    PSD_Fourier_domain(2:end,1,:)=(PSD_Fourier_domain(2:end,1,:)+PSD_Fourier_domain(end:-1:2,1,:))/2; % impose symmetry
    PSD_Fourier_domain(1,2:end,:)=(PSD_Fourier_domain(1,2:end,:)+PSD_Fourier_domain(1,end:-1:2,:))/2; % impose symmetry
    PSD_Fourier_domain(2:end,2:end,:)=(PSD_Fourier_domain(2:end,2:end,:)+PSD_Fourier_domain(end:-1:2,end:-1:2,:))/2; % impose symmetry

  
    %     PSD_Fourier_domain = abs(PSD_Fourier_domain); % % impose non-negativity (it's a PSD!)
    PSD_Fourier_domain = max(0,PSD_Fourier_domain); % % impose non-negativity (it's a PSD!)
    kerb=ones(5);kerb=kerb/sum(kerb(:));
    for juu=1:2
        PSD_Fourier_domain=convn(padarray(PSD_Fourier_domain,(size(kerb)-1)/2,'both','circular'),kerb,'valid');
    end
 
    % projecton onto subspace of desirable 
end

PSD_Fourier_domain = PSD_Fourier_domain./numel(PSD_FFT_domain)^2;