function Tr_PSD = PSD_domain_convertor(Fourier_PSD, PSD_size, Tr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  PSD_domain_convertor converts the input PSD (i.e. "Fourier_PSD") in Fourier-domain
%  (global FFT domain) into the PSD of size "PSD_size x PSD_size" in "Tr" domain (i.e. "Tr_PSD")
%
%  FUNCTION INTERFACE:
%     Tr_PSD = PSD_domain_convertor(Fourier_PSD, PSD_size, Tr)
% 
%  INPUTS:
%     1) Fourier_PSD      (2D/3D array) : Fourier-domain PSD to be converted
%     2) PSD_size         (double)      : the size of output PSD
%     3) Tr               (cell/char)   : transform doamin of the output PSD 
%
%  OUTPUTS:
%     1) Tr_PSD           (2D/3D array) : converted PSD in Tr doamin
%
%
%  EXAMPLE:
%-------------------------------------------------------------------------
% N                    = 1023;                       % NxN: size of 2-D noise signal
% L                    = 8;                          % LxL: size of converted 2-D PSD
% w                    = randn(N);                   % white Gaussian noise of size NxN
% g1                   = fspecial('gaussian',7,1.5); % lowpass Gaussian filter 
% g2                   = fspecial('gaussian',7,0.5); % lowpass Gaussian filter
% g                    = g1-g2;                      % correlation kernel (bandpass Gaussian filter)
% eta                  = imfilter(w,g,'circular');   % correlated noise
% Fourier_PSD          = abs(fft2(g,N,N)).^2*(N^2);  % noise PSD in Fourier (global FFT) domain    
% DCT_rootPSD_eta_cnvt = sqrt(PSD_domain_convertor(...
%    Fourier_PSD, [L L], {'dct', 'dct'}));           % DCT root-PSD via conversion
% DCT_rootPSD_eta_est  = Est_rootPSD(eta, [L L],...
%    {'dct','dct'}, 0, 'sVar');                      % DCT root-PSD via estimation
% 
% figure, subplot(1,2,1), imagesc(DCT_rootPSD_eta_cnvt), colormap('jet'),
% title('DCT root-PSD via conversion')
%         subplot(1,2,2), imagesc(DCT_rootPSD_eta_est), colormap('jet'),
% title('DCT root-PSD via estimation')
%-------------------------------------------------------------------------
%
%               AUTHORS:  Alessandro Foi & Nasser Eslahi
%                 contact: firstname.lastname@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2019 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if ~exist('PSD_size','var')||isempty(PSD_size)
    PSD_size        = [8 8];
end

if ~exist('Tr','var')||isempty(Tr)
    if numel(rPSD_size)==2
        Tr          = {'dct', 'dct'};         
    else
        Tr          = {'dct', 'dct', 'dct'}; 
    end 
end

s1                  = PSD_size(1);
s2                  = PSD_size(2);
if numel(PSD_size)==2
    s3              = 1;
else
    s3              = PSD_size(3);
end

TM_1                = single( GetTransformMatrix( s1, Tr{1,1} ) );
TM_2                = single( GetTransformMatrix( s2, Tr{1,2} ) );

if s3~=1
    TM_3            = single( GetTransformMatrix( s3, Tr{1,3} ) );
else
    TM_3            = 1;
end

[N1, N2, N3]        = size(Fourier_PSD);

Basis1              = zeros(N1,s1);
for j1 = 1:s1
    UnitMatrix      = zeros(s1,1); UnitMatrix(j1) = 1;
    Basis1(1:s1,j1) = TM_1'*UnitMatrix;  
end
Phi2_1              = abs(fft(Basis1)).^2/N1^2;   
clear Basis1 UnitMatrix


Basis2              = zeros(N2,s2);
for j2 = 1:s2
    UnitMatrix      = zeros(s2,1); UnitMatrix(j2)=1;
    Basis2(1:s2,j2) = TM_2'*UnitMatrix; 
end
Phi2_2              = abs(fft(Basis2)).^2/N2^2;  
clear Basis2 UnitMatrix


Basis3              = zeros(N3,s3);
for j3 = 1:s3
    UnitMatrix      = zeros(s3,1); UnitMatrix(j3)=1;
    Basis3(1:s3,j3) = TM_3'*UnitMatrix; 
end
Phi1_3              = reshape(abs(fft(Basis3)).^2,N3,s3);
clear Basis3 UnitMatrix

Tr_PSD              = zeros(s1,s2,N3);
for i = 1:N3
    Tr_PSD(:,:,i)   = Phi2_1'*Fourier_PSD(:,:,i)*Phi2_2;
end
%%% quivalent to the following (when s3=1):
% % for ii = 1:s1
% %     for jj = 1:s2
% %         UnitMatrix         = zeros(PSD_size); 
% %         UnitMatrix(ii,jj)  = 1;
% %         BasisElementPadded = zeros(N1, N2); 
% %         BasisElementPadded(1:s1,1:s2) = conj(TM_2')*UnitMatrix*TM_1;
% %         TransfBasisElementPadded      = fft2(BasisElementPadded);
% %         Tr_PSD(ii,jj)    = sum(sum(abs(TransfBasisElementPadded).^2.*Fourier_PSD))/(N1*N2)^2;
% %     end
% % end

Tr_PSD              = reshape(Tr_PSD,[s1*s2 N3]);
Tr_PSD              = Tr_PSD*Phi1_3./N3^2;
Tr_PSD              = reshape(Tr_PSD,[s1 s2 s3]);

end

%% Auxiliary function for creating the sparsifying transform basis used internally by the filter

function [Tforward, Tinverse] = GetTransformMatrix(N, transform_type, dec_levels)
%
% Create forward and inverse transform matrices, which allow for perfect
% reconstruction. The forward transform matrix is normalized so that the
% l2-norm of each basis element is 1.
%
% [Tforward, Tinverse] = getTransfMatrix (N, transform_type, dec_levels)
%
%  INPUTS:
%
%   N               --> Size of the transform (for wavelets, must be 2^K)
%
%   transform_type  --> 'dct', 'dst', 'hadamard', or anything that is
%                       listed by 'help wfilters' (bi-orthogonal wavelets)
%                       'DCrand' -- an orthonormal transform with a DC and all
%                       the other basis elements of random nature
%
%   dec_levels      --> If a wavelet transform is generated, this is the
%                       desired decomposition level. Must be in the
%                       range [0, log2(N)-1], where "0" implies
%                       full decomposition.
%
%  OUTPUTS:
%
%   Tforward        --> (N x N) Forward transform matrix
%
%   Tinverse        --> (N x N) Inverse transform matrix
%

if exist('dec_levels','var') ~= 1
    dec_levels = 0;
end

if N == 1
    Tforward = 1;
elseif strcmpi(transform_type, 'hadamard') == 1
    Tforward    = hadamard(N);
elseif (N == 8) && strcmpi(transform_type, 'bior1.5')==1 % hardcoded transform so that the wavelet toolbox is not needed to generate it
    Tforward =  [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.219417649252501   0.449283757993216   0.449283757993216   0.219417649252501  -0.219417649252501  -0.449283757993216  -0.449283757993216  -0.219417649252501;
        0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846  -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284;
        -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284   0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846;
        0.707106781186547  -0.707106781186547                   0                   0                   0                   0                   0                   0;
        0                   0   0.707106781186547  -0.707106781186547                   0                   0                   0                   0;
        0                   0                   0                   0   0.707106781186547  -0.707106781186547                   0                   0;
        0                   0                   0                   0                   0                   0   0.707106781186547  -0.707106781186547];
elseif (N == 8) && strcmpi(transform_type, 'dct')==1 % hardcoded transform so that the signal processing toolbox is not needed to generate it
    Tforward = [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.490392640201615   0.415734806151273   0.277785116509801   0.097545161008064  -0.097545161008064  -0.277785116509801  -0.415734806151273  -0.490392640201615;
        0.461939766255643   0.191341716182545  -0.191341716182545  -0.461939766255643  -0.461939766255643  -0.191341716182545   0.191341716182545   0.461939766255643;
        0.415734806151273  -0.097545161008064  -0.490392640201615  -0.277785116509801   0.277785116509801   0.490392640201615   0.097545161008064  -0.415734806151273;
        0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274   0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274;
        0.277785116509801  -0.490392640201615   0.097545161008064   0.415734806151273  -0.415734806151273  -0.097545161008064   0.490392640201615  -0.277785116509801;
        0.191341716182545  -0.461939766255643   0.461939766255643  -0.191341716182545  -0.191341716182545   0.461939766255643  -0.461939766255643   0.191341716182545;
        0.097545161008064  -0.277785116509801   0.415734806151273  -0.490392640201615   0.490392640201615  -0.415734806151273   0.277785116509801  -0.097545161008064];
elseif (N == 8) && strcmpi(transform_type, 'dst')==1 % hardcoded transform so that the PDE toolbox is not needed to generate it
    Tforward = [ 0.161229841765317   0.303012985114696   0.408248290463863   0.464242826880013   0.464242826880013   0.408248290463863   0.303012985114696   0.161229841765317;
        0.303012985114696   0.464242826880013   0.408248290463863   0.161229841765317  -0.161229841765317  -0.408248290463863  -0.464242826880013  -0.303012985114696;
        0.408248290463863   0.408248290463863                   0  -0.408248290463863  -0.408248290463863                   0   0.408248290463863   0.408248290463863;
        0.464242826880013   0.161229841765317  -0.408248290463863  -0.303012985114696   0.303012985114696   0.408248290463863  -0.161229841765317  -0.464242826880013;
        0.464242826880013  -0.161229841765317  -0.408248290463863   0.303012985114696   0.303012985114696  -0.408248290463863  -0.161229841765317   0.464242826880013;
        0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863;
        0.303012985114696  -0.464242826880013   0.408248290463863  -0.161229841765317  -0.161229841765317   0.408248290463863  -0.464242826880013   0.303012985114696;
        0.161229841765317  -0.303012985114696   0.408248290463863  -0.464242826880013   0.464242826880013  -0.408248290463863   0.303012985114696  -0.161229841765317];
elseif strcmpi(transform_type, 'dct') == 1
    Tforward    = dct(eye(N));
elseif strcmpi(transform_type, 'fft') == 1
    Tforward    = fft(eye(N));    
elseif strcmpi(transform_type, 'dst') == 1
    Tforward    = dst(eye(N));
elseif strcmpi(transform_type, 'eye') == 1
    Tforward    = eye(N);    
elseif strcmpi(transform_type, 'DCrand') == 1
    x = randn(N); x(1:end,1) = 1; [Q,~] = qr(x);
    if (Q(1) < 0)
        Q = -Q;
    end
    Tforward = Q';
else %% a wavelet decomposition supported by 'wavedec'
    %%% Set periodic boundary conditions, to preserve bi-orthogonality
    dwtmode('per','nodisp');
    
    Tforward = zeros(N,N);
    for i = 1:N
        Tforward(:,i)=wavedec(circshift([1 zeros(1,N-1)],[dec_levels i-1]), log2(N), transform_type);  %% construct transform matrix
    end
end


%%% Normalize the basis elements
if strcmpi(transform_type, 'fft') == 0
    Tforward = (Tforward' * diag(sqrt(1./sum(Tforward.^2,2))))';
end

%%% Compute the inverse transform matrix
Tinverse = inv(Tforward);

end
