function [A, At, S] = Sampling(im, L, Trajectory, seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  FUNCTION INTERFACE:
%   [A, At, S] = Sampling(im, L, Trajectory, seed)
%
% INPUTS:
%   im:               (double)          Input 2D/3D image.
%   L:                (double)          radial line numbers in the case pseudo-radial line 
%                                       or limited-angle sampling; or measurement ratio in 
%                                       the case of pseudo-random sampling.
%   Trajectory:       (string)          'radial' : radial line sampling
%                                       'lim_angle': limited angle sampling 
%                                       'pseudo_random': psueod random sampling
%                                       '3D_radial' : 3D radial line sampling.
%
% OUTPUTS:
%   A:                (function_handle) forward operator (sampling in FFT domain).
%   At:               (function_handle) inverse operator of "A".
%   S:                (logic)           sampling indicator.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if nargin == 3
    seed  = 0;
end

if     strcmpi(Trajectory,'radial')==1 ||...
       strcmpi(Trajectory,'lim_angle')==1     % radial line or limited angel sampling
    [A, At, S]      =   Radial_Line(im, L, seed, Trajectory );
elseif strcmpi(Trajectory,'pseudo_random')==1 % pseudo-random sampling
    [A, At, S]      = Pseudo_Random(im, L, seed);
elseif strcmpi(Trajectory,'3D_radial')==1     % 3D radial line sampling
    [A, At, S] =  Volumetric_Radial_Line( im, L, seed );
else
    error('check your sampling trajectory')
end


return;

          

function [A, At, S] =  Radial_Line( im, L, seed, Trajectory )
rng(seed,'v4');
[h, w]  =    size(im);

if strcmpi(Trajectory,'radial')==1
    aperture  = (pi/180)*180;    % aperture encompassing all scanning angles (aperture<pi is limited angle)
    direction = (pi/180)*0;     % direction of the scanning beam (middle axis)
else
    aperture  = (pi/180)*90;     % aperture encompassing all scanning angles (aperture<pi is limited angle)
    direction = (pi/180)*45;    % direction of the scanning beam (middle axis)    
end
S       =    LineMaskLimitedAngle(L,h,w,aperture,direction,seed);
A       =    @(z) S.*fft2(z);
At      =    @(z) ifft2(z);
return



function S = LineMaskLimitedAngle(L,h,w,aperture,direction,seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LineMaskLimitedAngle.m
%
% Returns the indicator of the domain in 2D Fourier space for the
% specified line geometry.
% Usage :  [S,Mh,mi,mhi] = LineMaskLimitedAngle(L,n,aperture,direction)
%
% n : size of domain
% L : number of lines
% aperture : aperture of fan
% direction : central direction of fan
%
% Alessandro Foi, Tampere University of Technology (2006-2011)
%
% Based on the original file LineMask.m written by Justin Romberg (rev. 12/2/2004) 
% as included in the L1magic package (http://www.acm.caltech.edu/l1magic/)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
if nargin<3
    aperture=pi;
end
if nargin<4
    direction=0;
end

if (pi-aperture)>(aperture/L)
    thc = linspace(-direction-aperture/2, -direction+aperture/2, L);
else
    thc = linspace(-direction-pi/2, -direction+pi/2-pi/L, L);
end

thc=mod(thc,pi);

S = zeros(h,w);
for ll = 1:L
    if ((thc(ll) <= pi/4) || (thc(ll) > 3*pi/4))
        yr        = round(tan(thc(ll))*(-w/2+1:w/2-1)+h/2+1);
        yr(yr<=0) = NaN;
        yr(yr>h)  = NaN;
        for nn = 1:w-1
            if ~isnan(yr(nn))
                S(yr(nn),nn+1) = 1;
            end
        end
    else
        xc        = round(cot(thc(ll))*(-h/2+1:h/2-1)+w/2+1);
        xc(xc<=0) = NaN;
        xc(xc>w)  = NaN;
        for nn = 1:h-1
            if ~isnan(xc(nn))
                S(nn+1,xc(nn)) = 1;
            end
        end
    end
end

S = ifftshift(S);    % sampling mask
return



function [A, At, S]   =  Pseudo_Random( im, ratio, seed )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Written by Weisheng Dong 
% http://see.xidian.edu.cn/faculty/wsdong/NLR_Exps.htm
% W. Dong, G. Shi, X. Li, Y. Ma, and F. Huang, "Compressive sensing via 
% Nonlocal Low-rank Regularization," IEEE Trans. Image Process., vol. 23, 
% no. 8, pp. 3618-3632, 2014.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(seed,'v4');
[h, w]       =    size(im);
if ratio==0.2
    factor   =    4.427;
elseif ratio==0.25
    factor   =    4;
else
    factor   =    sqrt(1/ratio)*2;
end
picks        =    RandMask_rect( double(h/factor), double(w/factor), h, w);
S            =    zeros(h*w,1);
S(picks)     =    1;
S            =    reshape(S,[h w]);   % sampling mask
A            =    @(z) S.*fft2(z);
At           =    @(z) ifft2(z);

return



function [mhi] = RandMask_rect(tolm, toln, m, n)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Shiqian Ma, Wotao Yin, Yin Zhang and Amit Chakraborty, "An Efficient Algorithm 
% for Compressed MR Imaging Using Total Variation and Wavelets", in IEEE CVPR 2008. 
% Author: Shiqian Ma
% Date : 09/05/2007
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(0,'v4');
M = zeros(m,n);

for i=1:m
	for j=1:n
		if ( ( abs(round(i-m/2)-(rand-0.5)*m/2)  < tolm ) &&  ( abs(round(j-n/2)-(rand-0.5)*n/2)  < toln ) )
			M(i,j) = 1;
		end
	end
end

cm         = round(m/2);
cn         = round(n/2);
M(cm,cn)   = 1;
M          = ifftshift(M);
M(1)       = 1;
mhi        = find(M);
return



function [A, At, S] =  Volumetric_Radial_Line( im, coverage, seed )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Returns the indicator of the domain in 3D Fourier space for the 3D radial
% line sampling.
% M. Maggioni, V. Katkovnik, K. Egiazarian, A. Foi, "A Nonlocal Transform-Domain Filter 
% for Volumetric Data Denoising and Reconstruction", IEEE Trans. Image Process.,
% vol. 22, no. 1, pp. 119-133, January 2013. doi:10.1109/TIP.2012.2210725
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
coverage = coverage/100;       % coverage of the 3D FFT space for 3D radial line sampling
tol      = 1/100;          % tolerance error of coverage 
line_std = 0/100;          % noise in subsampling trajectories
history  = zeros(2);
stop     = 0;
S        = zeros(size(im));
n        = size(im,1);
scale    = 8;
p0       = [0 0 0];
p1       = size(S);
t = 0:1/n:1;
while (sum(S(:))==0 || abs(mean(S(:))-coverage)>tol) && ~stop
    if mean(S(:))>coverage
        scale = scale+1;
    else
        scale = scale-1;
    end
    if scale==history(1,1)
        scale = history(1,history(2,:)==min(history(2,:)));
        stop = 1;
    end
    S = zeros(size(im));
    for sx=0:scale:p1(2)
        for sy=0:scale:p1(1)
            for sz=0:scale:p1(3)
                a = p0;
                b = p1;
                a(1) = max(1,min(p1(2),a(1)+sx));
                a(2) = max(1,min(p1(1),a(2)+sy));
                a(3) = max(1,min(p1(3),a(3)+sz));
                
                b(1) = max(1,min(p1(2),b(1)-sx));
                b(2) = max(1,min(p1(1),b(2)-sy));
                b(3) = max(1,min(p1(3),b(3)-sz));
                d = b-a;
                x = max(1,min(p1(2),round(a(1)+d(1).*t+line_std*randn)));
                y = max(1,min(p1(1),round(a(2)+d(2).*t+line_std*randn)));
                z = max(1,min(p1(3),round(a(3)+d(3).*t+line_std*randn)));
                for i=1:length(t)
                    S(y(i),x(i),z(i)) = 1;
                end
            end
        end
    end
    history = circshift(history,[0 1]);
    history(1,end) = scale;
    history(2,end) = abs(mean(S(:))-coverage);
end
for i=1:size(S,3)
    S(:,:,i) = ifftshift(S(:,:,i));
end
for i=1:size(S,1)
    for j=1:size(S,2)
        S(i,j,:) = squeeze(ifftshift(S(i,j,:)));
    end
end

A                 = @(z) S.*fftn(z);
At                = @(z) ifftn(z);
return

